#import python modules
import numpy as np
import time

#import compiled fortran library
import exfinallib

#import pdb writing
import atomwrite

#import os for deleting data file
import os


#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis
    
#whether or not to save the trajectory as a PdbFile
SaveTrajectory = True

#distance cutoff for pairwise interactions
Cut = 3.0


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    #make an array of lattice sites
    r = L * (np.arange(NLat, dtype=float)/NLat - 0.5)
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos, L, Box = False):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
    L: simulation box length
Output:
    Accel: (N,3) array of acceleration vectors
"""
    Accel = np.zeros_like(Pos)
    #get the acceleration from the forces
    if Box:
        PEnergy, Accel = exfinallib.energyforcesbox(Pos, L, Cut, Accel)
    else:
        PEnergy, Accel = exfinallib.energyforces(Pos, L, Cut, Accel)
    return Accel  



def InitBox():
    #set the init box width, number of particles, temperature, and timestep
    N = 108*5
    rho = (0.73*(0.33967)**3.)
    rho = 0.7041
    L = (N / rho)**(1./3.)
    Temp = 115.77 / 117.05
    dt = 0.001

    #set the frequency in seconds to update the display    
    DisplayFreq = 0.1
    #set the frequency in md steps to rescale velocities
    RescaleSteps = 1000
    #set the frequency in md steps to write coordinates
    WriteSteps = 100

    #set the max number of md steps; 0 for infinite loop
    MaxSteps = WriteSteps*1000

    #set the random number seed; useful for debugging
    #np.random.seed = 342324

    #get the initial positions, velocities, and acceleration (forces)    
    Pos = InitPositions(N, L)
    Vel = InitVelocities(N, Temp)
    Accel = InitAccel(Pos, L, Box = True)

    #set up the 3D display, if used    
    if UseVisual:
        atomvis.Clear()
        atomvis.Init(Pos, L)

    #MD steps
    StartTime = time.time()
    LastTime = time.time()
    i = 0
    if SaveTrajectory:
        Pdb = atomwrite.pdbfile("animfinalinit.pdb", L)
    while i < MaxSteps or MaxSteps <= 0:
        #do one step of the integration by calling the Fortran libraries
        Pos, Vel, Accel, KEnergy, PEnergy = exfinallib.vvintegratebox(Pos, Vel, Accel, L, Cut, dt)
        i += 1
        
        #check if we need to rescale the velocities 
        if i % RescaleSteps == 0:
            Vel = RescaleVelocities(Vel, Temp)

        #check if we need to output the positions 
        if SaveTrajectory and WriteSteps > 0 and i % WriteSteps == 0:
            Pdb.write(Pos)

        #check if we need to update the display            
        if time.time() - LastTime > DisplayFreq:
            print "%d  %11.4f  %11.4f  %11.4f" % (i, PEnergy + KEnergy, PEnergy, KEnergy)
            LastTime = time.time()
            
        #update the 3D visualization
        if UseVisual:
            atomvis.Update(Pos, L)

    if SaveTrajectory:
        Pdb.close()            

    #do one last update of the display         
    if UseVisual:
        atomvis.Update(Pos, L, Force = True)
    print "%d  %11.4f  %11.4f  %11.4f" % (i, PEnergy + KEnergy, PEnergy, KEnergy)

    StopTime = time.time()
    print "Total time: %.1f s" % (StopTime - StartTime)
    
    return Pos, N, L, Temp, dt


def Equil(Pos, N, L, Temp, dt):

    #set the frequency in seconds to update the display    
    DisplayFreq = 0.1
    #set the frequency in md steps to rescale velocities
    RescaleSteps = 1000
    #set the frequency in md steps to write coordinates
    WriteSteps = 100

    #set the max number of md steps; 0 for infinite loop
    MaxSteps = WriteSteps*1000

    #set the random number seed; useful for debugging
    #np.random.seed = 342324

    #get the initial positions, velocities, and acceleration (forces)    
    Pos = Pos % L
    Pos[:,2] = Pos[:,2] + L
    Vel = InitVelocities(N, Temp)
    Accel = InitAccel(Pos, L)

    #set up the 3D display, if used    
    if UseVisual:
        atomvis.Clear()
        atomvis.Init(Pos, [L, L, L*3.])
    

    #MD steps
    StartTime = time.time()
    LastTime = time.time()
    i = 0
    if SaveTrajectory:
        Pdb = atomwrite.pdbfile("animfinalequil.pdb", [L, L, L*3.])
    while i < MaxSteps or MaxSteps <= 0:
        #do one step of the integration by calling the Fortran libraries
        Pos, Vel, Accel, KEnergy, PEnergy = exfinallib.vvintegrate(Pos, Vel, Accel, L*3., Cut, dt)
        i += 1
        
        #check if we need to rescale the velocities 
        if i % RescaleSteps == 0 and False: ################# ADDED FALSE TO TEST #############################
            Vel = RescaleVelocities(Vel, Temp)

        #check if we need to output the positions 
        if SaveTrajectory and WriteSteps > 0 and i % WriteSteps == 0:
            Pdb.write(Pos)            

        #check if we need to update the display            
        if time.time() - LastTime > DisplayFreq:
            print "%d  %11.4f  %11.4f  %11.4f" % (i, PEnergy + KEnergy, PEnergy, KEnergy)
            LastTime = time.time()
            
        #update the 3D visualization
        if UseVisual:
            atomvis.Update(Pos, [L, L, L*3.])

    if SaveTrajectory:
        Pdb.close()            

    #do one last update of the display         
    if UseVisual:
        atomvis.Update(Pos, L, Force = True)
    print "%d  %11.4f  %11.4f  %11.4f" % (i, PEnergy + KEnergy, PEnergy, KEnergy)

    StopTime = time.time()
    print "Total time: %.1f s" % (StopTime - StartTime)
    
    return Pos, Vel, Accel

def Prod(Pos, N, L, Temp, dt, Vel, Accel):

    #set the frequency in seconds to update the display    
    DisplayFreq = 0.1
    #set the frequency in md steps to rescale velocities
    RescaleSteps = 1000
    #set the frequency in md steps to write coordinates
    WriteSteps = 100
    #Set the frequency in md steps to calculate things
    CalcSteps = 100

    #set the max number of md steps; 0 for infinite loop
    MaxSteps = WriteSteps*2500

    #set the random number seed; useful for debugging
    #np.random.seed = 342324

    #get the initial positions, velocities, and acceleration (forces)    

#    Pos, Vel, Accel, KEnergy, PEnergy = exfinallib.vvintegrate(Pos, Vel, Accel, L*3., Cut, 0)
#    print "%d  %11.4f  %11.4f  %11.4f" % (i, PEnergy + KEnergy, PEnergy, KEnergy)
    #Pos[:,2] = Pos[:,2] - L
    #Pos = Pos % np.array([L, L, 3.*L])
    #Vel = InitVelocities(N, Temp)   #Create Fake Velocities
    #Accel = InitAccel(Pos, L)       #Create Fake Acceleration
    #Pos, Vel, Accel, KEnergy, PEnergy = exfinallib.vvintegrate(Pos, Vel, Accel, L*3., Cut, 0)   #Calculate true values for velocity and acceleration

    #set up the 3D display, if used    
    if UseVisual:
        atomvis.Clear()
        atomvis.Init(Pos, [L, L, L*3.])

    #Test to see if the data file exists
    Is_Taken = True
    num = 1
    while Is_Taken:
        try: Is_Taken = os.path.exists('exfinaldata%03i.txt' %num)
        except: Is_Taken = True
        DataFile = 'exfinaldata%03i.txt' %num
        print('Try Number: %i' %num)
        print(DataFile)
        print(Is_Taken)
        num += 1
    data = file(DataFile,'a')

    #MD steps
    StartTime = time.time()
    LastTime = time.time()
    i = 0
    if SaveTrajectory:
        Pdb = atomwrite.pdbfile("animfinalprod.pdb", [L, L, L*3.])
    while i < MaxSteps or MaxSteps <= 0:
        #do one step of the integration by calling the Fortran libraries
        Pos, Vel, Accel, KEnergy, PEnergy = exfinallib.vvintegrate(Pos, Vel, Accel, L*3., Cut, dt)
        i += 1
        
        #check if we need to rescale the velocities 
        #if i % RescaleSteps == 0 and False: ################# ADDED FALSE TO TEST #############################
        #    Vel = RescaleVelocities(Vel, Temp)
        #Do Not Rescale Velocities in Production Run!!!!!!!!!!!!!

        #check if we need to output the positions 
        if SaveTrajectory and WriteSteps > 0 and i % WriteSteps == 0:
            Pdb.write(Pos)            

        #check if we need to update the display            
        if time.time() - LastTime > DisplayFreq:
            print "%d  %11.4f  %11.4f  %11.4f" % (i, PEnergy + KEnergy, PEnergy, KEnergy)
            LastTime = time.time()
            
        #update the 3D visualization
        if UseVisual:
            atomvis.Update(Pos, [L, L, L*3.])
        
        if i % CalcSteps == 0:
            Pos_hist = Pos % np.array([L, L, 3.*L])
            #Pos_hist = Pos_hist[Pos_hist[:,2].argsort()]
            hist = np.zeros(300)
            for each in range(len(Pos)):
                #print('Here is the index Im trying to access')
                #print(int(300*Pos_hist[each][2] / L))
                #print('Here is the value of Pos_hist at %i: %0.5f' %(each, Pos_hist[each][2]))
                hist[int(300*Pos_hist[each][2] / (3.*L))] += 1
            data.write('%i\t' %i)
            for each in range(len(hist)):
                data.write('%i\t' %hist[each])
            data.write('\n')
                
                

    if SaveTrajectory:
        Pdb.close()
    data.close()

    #do one last update of the display         
    if UseVisual:
        atomvis.Update(Pos, L, Force = True)
    print "%d  %11.4f  %11.4f  %11.4f" % (i, PEnergy + KEnergy, PEnergy, KEnergy)

    StopTime = time.time()
    print "Total time: %.1f s" % (StopTime - StartTime)
    
    return Pos

#check to see if we were run at the command line
if __name__ == '__main__':
    #run the test simulation
    Pos, N, L, Temp, dt = InitBox()
    Pos, Vel, Accel = Equil(Pos, N, L, Temp, dt)
    Pos = Prod(Pos, N, L, Temp, dt, Vel, Accel)
